<?php
/**
 * EAdvancedRelations class file.
 *
 * @author Lucas Calje <lucas@calje.eu>
 * @link http://www.yiiframework.com/
 * @version 0.1
 */

 /**
  * The EAdvancedArBehavior extension takes away some of the complexities when saving object, for example, with MANY_MANY and BELONGS_TO relations.
  * 
  * 
  * Usage
  *
  * The EAdvancedArBehavior extension adds up some functionality to the default possibilites of yiis ActiveRecord implementation.
  * To use this extension, just copy this file to your extensions/ directory, add 'import' => 'application.extensions.EAdvancedArBehavior', [...] to your config/main.php and add this behavior to each model you would like to inherit the new possibilities:
  *
  * public function behaviors(){
       * return array( 'EAdvancedArBehavior' => array(
             * 'class' => 'application.extensions.EAdvancedArBehavior'));
  * }
  *
  * 
  *
  * Better support of MANY_TO_MANY relations: 
  *
  * see CAdvancedArBehavior (http://www.yiiframework.com/extension/cadvancedarbehavior/)
  *
  *
  * Better support of BELONGS_TO relations:
  *
  * With this extension you can connect two related object as follows:
  * 
  * $user = new User();
  * $user->company = Company::model()->find($conditions, $params);
  * $post->save();
  * 
  * Without this extension, you would do this as follows:
  *
  * $user = new User();
  * $company = Company::model()->find($conditions, $params) ;
  * $user->company_id = $company->id ;
  * $post->save();
  *
  *
  * Set uninitialized column fields to NULL
  * 
  * See see: http://code.google.com/p/yiiext/downloads/detail?name=ensureNull_1.0.1.zip
  *
  * Use it like
  *
  * return array(
  * 	'EAdvancedArBehavior' => array(
  *     	'class' => 'application.extensions.EAdvancedArBehavior',
  *     	'useOnUpdate' => false, // set empty fields to NULL only on inserts not updates
  * 	);
  *
  *
  * MySQL ONLY: Multiple timespans with CURRENT_TIMESTAMP
  * 
  * Although the title of this section suggests something which cannot be done with MySQL, there is a nice
  * workaround available. This extension provides the means to have a table with one column for the insert time and an
  * other for the last modification time.
  * First, add the following two columns to your table definition
  * 
  * update_time timestamp NOT NULL DEFAULT '0000-00-00 00:00:00' ON UPDATE CURRENT_TIMESTAMP,
  * created_time timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  *
  * Next, configure the the following parameters
  * 
  * return array(
  *     'EAdvancedArBehavior' => array(
  *             'class' => 'application.extensions.EAdvancedArBehavior',
  *             'useNullOnTimestamp' => TRUE,         // set timestamps with a default value of 0000-00-00 00:00 to NULL (for inserts)
  *             'onUpdateTimestamp'=>'update_time',   // column name of the update time
  *             ),
  *       );
 */

class EAdvancedArbehavior extends CActiveRecordBehavior
{
	public $useOnUpdate			= TRUE ;	// set empty fields to NULL when updating (NOTE: on insert nullable empty attributes are set to NULL by default) 
	public $useNullOnTimestamp  = FALSE ; 	// if TRUE it will set timestamp values with a default value of $timestampDefault to NULL
	public $onUpdateColumn   	= NULL ;	// define the column-name holding the modification time
	public $timestampDefault 	= '0000-00-00 00:00:00' ; // default timestamp value
	
	public function afterSave($on) 
	{
		$this->fixRelations() ;
		$this->writeManyManyTables();
		return TRUE;
	}
	
	public function beforeValidate() { 
		$this->fixBELONGS_TO() ;
		return TRUE ;
	}

	public function fixBELONGS_TO() {
		$owner = $this->owner ;
		foreach($owner->relations() as $key => $relation)
                {
                        /* $key         -> relation name
			 * $relation[1] -> related table
                         * $relation[2] -> foreignkey field
                         */
			if($relation['0'] == CActiveRecord::BELONGS_TO) // ['0'] equals relationType
                        {
                                Yii::trace('set BELONGS_TO foreign-key field for '.get_class($owner),'system.db.ar.CActiveRecord');
                                if ( isset($owner->{$key}) ) {
                                        $owner->$relation[2] = $owner->{$key}->primaryKey ;
                                }
                        }
		}

	}
	public function fixRelations() {
		$owner = $this->owner ; 
		
		foreach($owner->relations() as $key => $relation)
                {
                        /* $key         -> relation name
			 * $relation[1] -> related table
                         * $relation[2] -> foreignkey field
                         */
			if($relation['0'] == CActiveRecord::HAS_ONE) // ['0'] equals relationType
                        {
				// update the related table which contains the foreignkey (BELONGS_TO)
                                Yii::trace('set HAS_ONE foreign-key field for '.get_class($owner),'system.db.ar.CActiveRecord');
				$related = $owner->{$key} ;
				if ( isset($related) && ( empty($related->{$relation[2]}) || $related->{$relation[2]} !=  $owner->primaryKey) ) {
					$related->{$relation[2]} = $owner->primaryKey ;
					$related->save() ;		
				}
			}
			else if ( $relation['0'] == CActiveRecord::HAS_MANY ) {
                                Yii::trace('set HAS_MANY foreign-key field for '.get_class($owner),'system.db.ar.CActiveRecord');
				$related = $owner->{$key} ; 
				if ( isset($related) ) {
					foreach ( $related as $obj ) {
						if ( empty($obj->{$relation[2]}) || $obj->{$relation[2]} !=  $owner->primaryKey ) {
							$obj->{$relation[2]} = $owner->primaryKey ;
							$obj->save(false) ;
						}
					}
				}
			}
		}
	}

	public function beforeSave($event)
	{
		$owner=$this->getOwner();

		if($owner->getIsNewRecord() || $this->useOnUpdate)
		{
			$mysqlOnly = $this->useNullOnTimestamp && $owner->dbConnection->driverName == 'mysql' ;
			
			foreach($owner->getTableSchema()->columns as $column)
			{
				$value = trim($owner->getAttribute($column->name)) ;
				if( ( $column->allowNull && $value==='')	// set nullable empty fields to NULL
					  || 
					( $mysqlOnly == 1 && ( $owner->isNewRecord && $value == $this->timestampDefault && $column->name != $this->onUpdateColumn	  // set insert-time field to NULL
										  ||
						 				  !$owner->isNewRecord && $column->name == $this->onUpdateColumn	// set update-time field to NULL
						 				)
					)
				) {
					$column->allowNull = TRUE ; 
					$owner->setAttribute($column->name,null);
				}
			}
		}
		return TRUE ;
	}

	/**
	 * At first, this function cycles through each MANY_MANY Relation. Then
	 * it checks if the attribute of the Object instance is an integer, an
	 * array or another ActiveRecord instance. It then builds up the SQL-Query
	 * to add up the needed Data to the MANY_MANY-Table given in the relation
	 * settings.
	 */
	public function writeManyManyTables() 
	{
		Yii::trace('writing MANY_MANY data for '.get_class($this->owner),'system.db.ar.CActiveRecord');

		foreach($this->owner->relations() as $key => $relation)
		{
			if($relation['0'] == CActiveRecord::MANY_MANY) // ['0'] equals relationType
			{
				if(isset($this->owner->$key))
				{ // MANY_MANY is set
					if(is_object($this->owner->$key) || is_numeric($this->owner->$key))
					{
						if ( is_numeric($this->owner->$key) || $this->findOwner($this->owner->$key) == FALSE ) {
							$this->executeManyManyEntry($this->makeManyManyDeleteCommand(
								$relation[2],
								$this->owner->{$this->owner->tableSchema->primaryKey}));
							$this->executeManyManyEntry($this->owner->makeManyManyInsertCommand(
								$relation[2],
								(is_object($this->owner->$key))
								? $this->owner->$key->{$this->owner->$key->tableSchema->primaryKey}
								: $this->owner->{$key}));
						}
					}
					else if (is_array($this->owner->$key) ) 
					{
						list( $toBeRemovedArray, $toBeRemovedHash) =  $this->loadOldRelationsFromDB($relation[2]) ;
						$diff = array_diff( $toBeRemovedArray, $this->convertObjsToPrimaryKeyArray($this->owner->{$key})) ;

						if ( sizeof($diff) > 0 )  {
							$this->executeManyManyEntry($this->makeManyManyDeleteCommand(
								$relation[2],
								$this->owner->primaryKey,
								$diff
							)) ; 
						}
						$sql = null ;
						foreach($this->owner->$key as $foreignobject)
						{
							if ( empty($toBeRemovedHash[$foreignobject->primaryKey]) ) { // new relation
								$sql = $this->makeManyManyInsertCommand(
									$relation[2],
									(is_object($foreignobject))
									? $foreignobject->{$foreignobject->tableSchema->primaryKey}
									: $foreignobject, $sql);
							}
						}
						if ( isset($sql) )
							$this->executeManyManyEntry( $sql ) ;
					} 
				}
			}
		}
	}

	private function loadOldRelationsFromDB($model) {
		$secondRelation = trim($this->getManyManySecondRelationName($model)) ;
		$reader = $this->owner->dbConnection->createCommand( 
			   sprintf('SELECT %s FROM %s WHERE %s = %d',
				$secondRelation,
				$this->getManyManyTable($model),
				$this->getManyManyFirstRelationName($model),
				$this->owner->primaryKey
			   )
			)->query() ;

		$listA = array() ;
		$listH = array() ;
		foreach( $reader as $row )  {
			$listA[] = $row[$secondRelation] ;
			$listH[''.$row[$secondRelation]] = TRUE ;
		}
		return array( $listA, $listH ) ;
	}

	private function convertObjsToPrimaryKeyArray($list) {
		$primaryKeys = array() ;
		foreach( $list as $obj ) 
			$primaryKeys[] = $obj->primaryKey ;
		return $primaryKeys ;
	}

	// We can't throw an Exception when this query fails, because it is possible
	// that there is not row available in the MANY_MANY table, thus execute()
	// returns 0 and the error gets thrown falsely.
	public function executeManyManyEntry($query) {
		if ( isset($query) )
			return $this->owner->dbConnection->createCommand($query)->execute();
	}

	// It is important to use insert IGNORE so SQL doesn't throw an foreign key
	// integrity violation
	public function makeManyManyInsertCommand($model, $rel, $sql=null) {
		if (isset($sql) )  // append to INSERT statement
			return sprintf( '%s,(%s, %s)', $sql, $this->owner->{$this->owner->tableSchema->primaryKey}, $rel )  ;
		else // create INSERT statement
			return sprintf("insert ignore into %s values ('%s', '%s')", $model,	$this->owner->{$this->owner->tableSchema->primaryKey}, $rel);
	}

	public function makeManyManyDeleteCommand($model, $rel, $removeIDs = null) {
		$sql = sprintf('delete ignore from %s where %s = \'%s\'', $this->getManyManyTable($model), $this->getManyManyFirstRelationName($model), $rel);
		if ( isset($removeIDs) && sizeof($removeIDs) > 0 ) 
			$sql .= sprintf( ' AND %s IN (\'%s\')', $this->getManyManySecondRelationName($model), implode("','", $removeIDs) ) ;
		else
			return null ;
		return $sql ;
	}

	public function getManyManyTable($model) {
		if (($ps=strpos($model, '('))!==FALSE)
		{
			return substr($model, 0, $ps);
		}
		else
			return $model;
	}

	/** 
	  * @param string $model  ( e.g: tbl_project_user_assignment(user_id, project_id) )
	  * @return string (e.g:  user_id )
	  */
	public function getRelationNameForDeletion($model) {
		preg_match('/\((.*),/',$model, $matches) ;
		return substr($matches[0], 1, strlen($matches[0]) - 2);
	}

	public function getManyManyFirstRelationName($model) {
		return $this->getRelationNameForDeletion($model) ;
	}

	public function getManyManySecondRelationName($model) {
		preg_match('/,(.*?)\)/', $model, $matches) ;
		return $matches[1] ;
	}

	public function beforeDelete($event)
	{
    		Yii::trace('deleting MANY_MANY data for '.get_class($this->owner),'system.db.ar.CActiveRecord');
 	
    		foreach($this->owner->relations() as $key => $relation)
    		{
        		if($relation['0'] == CActiveRecord::MANY_MANY) // ['0'] equals relationType
        		{
                      		$this->executeManyManyEntry($this->makeManyManyDeleteCommand(
                              		$relation[2],
                              		$this->owner->{$this->owner->tableSchema->primaryKey}));
                    	}
            	}
    	}
}
